module 'data.number' {
   syntax: prefix '++' as '++|'.
   syntax: prefix '--' as '--|'.
   syntax: postfix '++' as '|++'.
   syntax: postfix '--' as '|--'.

   syntax: '>>' < infix '+'.
   syntax: prefix '-' < prefix '++' = prefix '--' = postfix '++'
      = postfix '--' < ':'.

   syntax: infix '..'.

   syntax: '=' < '..' < '+'.

   /** check(X).
    *  Succeeds if <code>X</code> is instantiated to a number.
    */

   check(X) :-
      'kernel':number_check(X).

   /** A >= B.
    *  <prerequisite>
    *  <code>A</code> and <code>B</code> are instantiated to numbers.
    *  </prerequisite>
    *  Succeeds if <math><code>A</code>&ge;<code>B</code></math>.
    */

   `>='(A, B) :-
      'kernel':number_ge(A, B).

   `>='(A, B, B) :-
      'kernel':number_ge(A, B).

   /** A > B.
    *  <prerequisite>
    *  <code>A</code> and <code>B</code> are instantiated to numbers.
    *  </prerequisite>
    *  Succeeds if <math><code>A</code>&gt;<code>B</code></math>.
    */

   `>'(A, B) :-
      'kernel':number_gt(A, B).

   `>'(A, B, B) :-
      'kernel':number_gt(A, B).

   /** A =< B.
    *  <prerequisite>
    *  <code>A</code> and <code>B</code> are instantiated to numbers.
    *  </prerequisite>
    *  Succeeds if <math><code>A</code>&le;<code>B</code></math>.
    */

   `=<'(A, B) :-
      'kernel':number_le(A, B).

   `=<'(A, B, B) :-
      'kernel':number_le(A, B).

   /** A < B.
    *  <prerequisite>
    *  <code>A</code> and <code>B</code> are instantiated to numbers.
    *  </prerequisite>
    *  Succeeds if <math><code>A</code>&lt;<code>B</code></math>.
    */

   `<'(A, B) :-
      'kernel':number_lt(A, B).

   `<'(A, B, B) :-
      'kernel':number_lt(A, B).

   /** pred(A, R).
    *  <prerequisite>
    *  <code>A</code> is instantiated to a number.
    *  </prerequisite>
    *  Succeeds if <code>R</code> is <math><code>A</code> - 1</math>.
    */

   pred(A, B) :-
      (
         kernel:number_check(B)
      ->
         kernel:number_succ(B, A)
      ;
         kernel:number_pred(A, B)
      ).

   /** succ(A, R).
    *  <prerequisite>
    *  <code>A</code> is instantiated to a number.
    *  </prerequisite>
    *  Succeeds if <code>R</code> is <math><code>A</code> + 1</math>.
    */

   succ(A, B) :-
      (
         kernel:number_check(B)
      ->
         kernel:number_pred(B, A)
      ;
         kernel:number_succ(A, B)
      ).

   /** + A = A.
    *  Identity function.
    */

   `+'(A) = A.

   /** - A = R.
    *  <prerequisite>
    *  <code>A</code> is instantiated to a number.
    *  </prerequisite>
    *  Succeeds if <code>R</code> is <code>-A</code>.
    */

   `-'(A, B) :-
      (
         kernel:number_check(B)
      ->
         'kernel':number_opp(B, A)
      ;
         'kernel':number_opp(A, B)
      ).

   /** A + B = R.
    *  <prerequisite>
    *  <code>A</code> and <code>B</code> are instantiated to numbers.
    *  </prerequisite>
    *  Succeeds if <code>R</code> is <math><code>A</code> + <code>B</code></math>.
    */

   `+'(A, B, C) :-
      (
         kernel:number_check(C)
      ->
         (
            kernel:number_check(B)
         ->
            kernel:number_sub(C, B, A)
         ;
            kernel:number_sub(C, A, B)
         )
      ;
         kernel:number_add(A, B, C)
      ).

   /** A - B = R.
    *  <prerequisite>
    *  <code>A</code> and <code>B</code> are instantiated to numbers.
    *  </prerequisite>
    *  Succeeds if <code>R</code> is <math><code>A</code> - <code>B</code></math>.
    */

   `-'(A, B, C) :-
      (
         kernel:number_check(C)
      ->
         (
            kernel:number_check(B)
         ->
            kernel:number_add(C, B, A)
         ;
            kernel:number_add(C, A, B)
         )
      ;
         kernel:number_sub(A, B, C)
      ).

   /** A * B = R.
    *  <prerequisite>
    *  <code>A</code> and <code>B</code> are instantiated to numbers.
    *  </prerequisite>
    *  Succeeds if <code>R</code> is <math><code>A</code> * <code>B</code></math>.
    */

   `*'(A, B, C) :-
      (
         kernel:number_check(C)
      ->
         (
            kernel:number_check(B)
         ->
            kernel:number_div(C, B, A)
         ;
            kernel:number_div(C, A, B)
         )
      ;
         kernel:number_mul(A, B, C)
      ).

   /** A / B = R.
    *  <prerequisite>
    *  <code>A</code> and <code>B</code> are instantiated to numbers.
    *  </prerequisite>
    *  Succeeds if <code>R</code> is <math><code>A</code> / <code>B</code></math>.
    */

   `/'(A, B, C) :-
      (
         kernel:number_check(C)
      ->
         (
            kernel:number_check(B)
         ->
            kernel:number_mul(C, B, A)
         ;
            kernel:number_mul(C, A, B)
         )
      ;
         kernel:number_div(A, B, C)
      ).

   /** A quo B = R.
    *  <prerequisite>
    *  <code>A</code> and <code>B</code> are instantiated to numbers.
    *  </prerequisite>
    *  Succeeds if <code>R</code> is the quotient in the euclidian division of
    *  <code>A</code> by <code>B</code>.
    */

   quo(A, B) = 'kernel':number_quo(A, B).

   /** A mod B = R.
    *  <prerequisite>
    *  <code>A</code> and <code>B</code> are instantiated to numbers.
    *  </prerequisite>
    *  Succeeds if <code>R</code> is the reminder in the euclidian division of
    *  <code>A</code> by <code>B</code>.
    */

   mod(A, B) = 'kernel':number_mod(A, B).

   max(A, B) = 'kernel':number_max(A, B).

   max([H | T]) = data.list:fold_left(T, H) {
      do(A, B) = max(A, B)
   }.

   /** A++.
    *  <prerequisite>
    *  <code>A</code> is a reference to an instantiated number.
    *  </prerequisite>
    *  Increments the number referenced by <code>A</code>.
    */

   `|++'(A) :-
      A:get(V),
      'data.number':succ(V, W),
      A:set(W).

   /** A++ = R.
    *  <prerequisite>
    *  <code>A</code> is a reference to an instantiated number.
    *  </prerequisite>
    *  Succeeds if <code>R</code> is the number referenced by <code>A</code>.
    *  In case of success, the referenced number is incremented.
    */

   `|++'(A, V) :-
      A:get(V),
      'data.number':succ(V, W),
      A:set(W).

   /** ++A = R.
    *  <prerequisite>
    *  <code>A</code> is a reference to an instantiated number.
    *  </prerequisite>
    *  Succeeds if <code>R</code> is the successor of the number referenced by
    *  <code>A</code>.  In case of success, the referenced number is incremented.  
    */

   `++|'(A, W) :-
      A:get(V),
      'data.number':succ(V, W),
      A:set(W).

   /** A--.
    *  <prerequisite>
    *  <code>A</code> is a reference to an instantiated number.
    *  </prerequisite>
    *  Decrements the number referenced by <code>A</code>.
    */

   `|--'(A) :-
      A:get(V),
      'data.number':pred(V, W),
      A:set(W).

   /** A-- = R.
    *  <prerequisite>
    *  <code>A</code> is a reference to an instantiated number.
    *  </prerequisite>
    *  Succeeds if <code>R</code> is the number referenced by <code>A</code>.
    *  In case of success, the referenced number is decremented.
    */

   `|--'(A, V) :-
      A:get(V),
      'data.number':pred(V, W),
      A:set(W).

   /** --A = R.
    *  <prerequisite>
    *  <code>A</code> is a reference to an instantiated number.
    *  </prerequisite>
    *  Succeeds if <code>R</code> is the predecessor of the number referenced by
    *  <code>A</code>.  In case of success, the referenced number is decremented.  
    */

   `--|'(A, V) :-
      A:get(V),
      'data.number':pred(V, W),
      A:set(W).

   /** atom(N, A).
    *  <prerequisite>
    *  <code>N</code> is a reference to an instantiated number.
    *  </prerequisite>
    *  Succeeds if <code>A</code> is the atom containing the decimal notation of
    *  the value of <code>N</code>.
    */

   atom(N, A) :-
      'kernel':number_to_atom(N, A).

   /** iter(I, J, F).
    *  <prerequisite>
    *  <code>I</code> and <code>J</code> are references to instantiated numbers.
    *  </prerequisite>
    *  <prerequisite>
    *  <code>F</code> is instantiated to a module which implements a predicate
    *  <code>do/1</code>.
    *  </prerequisite>
    *  Iterates <code>do(I + X)</code> for each integer <code>X</code> such that
    *  <math>0 &le; <code>X</code> &le; J - I</math>, increasingly.
    */

   iter(I, J, F) :-
      (
         'data.number':(I =< J)
      ->
         F:do(I),
         'data.number':iter('data.number':succ(I), J, F)
      ;
         true
      ).

   /** I .. J = List.
    *  <prerequisite>
    *  <code>I</code> and <code>J</code> are references to instantiated numbers.
    *  </prerequisite>
    *  Succeeds if <code>List</code> is the list <code>[I, I + 1, ..., I + K]</code>
    *  where <code>K</code> is the highest integer such that <code>I + K &le; J</code>.
    */

   `..'(I, J) =
      (
         I =< J
      ->
         [I | succ(I) .. J]
      ;
         []
      ).

   /** char(Code, Char).
    *  <prerequisite>
    *  <code>Code</code> is an instantiated code-point.
    *  </prerequisite>
    *  Succeeds if <code>Char</code> is the character, the code-point of which is
    *  <code>Code</code>.
    */

   char(Code, Char) :-
      'kernel':number_char(Code, Char).

   random(Generator, Value) :-
      Generator:random(Value).

   random(Generator, Base, Max, Value) :-
      Generator:random(Base, Max, Value).

   syntax: postfix '!'.

   syntax: prefix '-' < '!' < ':'. 

   `!'(N) = (N > 2, !) >> N * (N - 1)!.
   `!'(_) = 1.
}
