module emop.parse {
   constructor: callable/5, 'module'/7, number/3, empty/2, doc/3, none/0,
      error/2, operator/0, unescaped/0, atom/4.

   position(callable(_F, _Escaped, _Args, Pb, Pe), 'file'(Pb, Pe)).
   position('()'(_Term, Pb, Pe), 'file'(Pb, Pe)).
   position('.()'(_M, _Term, Pb, Pe), 'file'(Pb, Pe)).
   position(module(_Name, _Env, _Body, _Sgn, _Clauses, Pb, Pe), 'file'(Pb, Pe)).
   position(number(_, Pb, Pe), 'file'(Pb, Pe)).
   position(empty(Pb, Pe), 'file'(Pb, Pe)).
   position(doc(_, Pb, Pe), 'file'(Pb, Pe)).

   file(Filename, Grammar, Ast) :-
      io.file:open_read(Filename, Stream),
      stream(Filename, Stream, Grammar, Ast),
      Stream:close.

   stream(Filename, Stream, Grammar, Ast) :-
      emop.parser:parse_stream(Filename, 'file', Stream, _, Ast0),
      Ast = reparse(Ast0, Grammar).

   term(callable(F, _, Args, _, _)) =
      data.term:make(
         F,
         (Args = none -> []; data.list:iter(Args) { do(X) = term(X). })
      ).
   term('()'(Sub, _, _)) = term(Sub).
   term('.()'(_, _, _, _), _) :-
      throw(error('.()', 'term')).
   term(module(_, _, _, _, _, _, _), _) :-
      throw(error('module', 'term')).
   term(number(N, _, _)) = N.
   term(empty(_, _), _) :-
      throw(error('empty', 'term')).
   term(doc(_, _, _), _) :-
      throw(error('doc', 'term')).

   reparse(Ast0, Grammar, Ast1) :-
      reparse_flat(reflat(Ast0), Grammar, Ast1).

   reparse_flat(Ast0, Grammar, Ast1) :-
      Grammars = control.iter:findall() {
         do = Grammar:`$sgn'.
      },
      Grammars' = data.list:fold_left(Grammars, []) {
         do(Sgn, Acc) =
            [emop.parser:make_grammar(Sgn, Acc) | Acc].
      },
      emop.parser:reparse_flat(Ast0, Grammars', Ast1).

   reflat_append(callable('$~$', operator, [A, B], Pb0, Pe0), C, Pb, Pe) =
      ! >> reflat_append(A, reflat_append(B, C, Pb0, Pe0), Pb, Pe).
   reflat_append(A, B, Pb, Pe) =
      callable('$~$', operator, [A, B], Pb, Pe).

   reflat(callable('$~$', operator, [A, B], Pb, Pe)) =
      ! >> reflat_append(reflat(A), reflat(B), Pb, Pe).
   reflat(callable(F, operator, [A, B], Pb, Pe)) =
      ! >> reflat_append(
         reflat(A),
         callable(
            '$~$',
            operator,
            [
               callable(F, unescaped, none, Pb, Pe),
               reflat(B)
            ],
            Pb, Pe
         ),
         Pb, Pe
      ).
   reflat(callable(F, Escaped, Args, Pb, Pe)) =
      ! >> callable(F, Escaped, reflat_list(Args), Pb, Pe).
   reflat('()'(Sub, Pb, Pe)) =
      ! >> '()'(reflat(Sub), Pb, Pe).
   reflat('.()'(A, B, Pb, Pe)) =
      ! >> reflat_append(
         reflat(A),
         '.()'(B, Pb, Pe),
         Pb, Pe
      ).
   reflat(module(Name, Env, Body, Sgn, Clauses, Pb, Pe)) =
      ! >> module(Name, reflat_list(Env), Body, Sgn, Clauses, Pb, Pe).
   reflat(A) = A.

   reflat_list(none) = ! >> none.
   reflat_list(L) = data.list:iter(L) { do(X) = reflat(X) }.

   enum_concat_list(callable('$~$', operator, [A, B], _, _)) =
      ! >> [A | enum_concat_list(B)].
   enum_concat_list(A) = [A].

   decompose_pos('file'(Pb, Pe), Pb, Pe) :- !.
   decompose_pos(_, 'dummy', 'dummy').

   reconcat_list([A], _Pos) = ! >> A.
   reconcat_list([A | B], Pos) =
      decompose_pos(Pos, Pb, Pe) >>
      callable('$~$', operator, [A, reconcat_list(B, Pos)], Pb, Pe).


   fv(callable(Atom, Escaped, Args, _, _), V) :-
      !,
      (
         emop.parser:is_variable(Atom, Escaped, Args)
      ->
         (
            Atom = '_'
         ->
            V = []
         ;
            V = [Atom]
         )
      ;
         fv_list(Args, V)
      ).
   fv('()'(X, _, _), V) :-
      !,
      fv(X, V).
   fv('.()'(X, Y, _, _), V) :-
      !,
      fv(X, VX),
      fv(Y, VY),
      data.list:merge(VX, VY, 'data.term', V).
   fv(module(Name, Env, _, _, _, _, _), V) :-
      !,
      (
         Name = atom(Title, Escaped, _, _),
         emop.parser:is_variable(Title, Escaped, none),
         Title \= '_'
      ->
         V0 = [Title]
      ;
         V0 = []
      ),
      (
         Env = none
      ->
         V = V0
      ;
         V1 = data.list:morph(Env) {
            do(callable('_', _, _, _, _)) = ! >> [].
            do(callable(Atom, _, _, _, _)) = ! >> [Atom].
            do('()'(Sub, _, _)) = ! >> [Sub].
            do(_) = [].
         },
         V = data.list.(merge_sort(V0 ++ V1, 'data.term'))
      ).
   fv(_, []).

   fv_list(none, []).
   fv_list([], []).
   fv_list([H | T], V) :-
      fv(H, V0),
      fv_list(T, V1),
      data.list:merge(V0, V1, 'data.term', V).

}